#include <smmintrin.h>
#include <stdint.h>
#include <stdio.h> 
#include <time.h> 
#include <malloc.h>
#include "tsc_x86.h"

#define FLOATS_PER_INSTR 4

/* Define Infinite as a large enough value. This value will be used 
  for vertices not connected to each other */
#define INF 99999 
  
// A function to print the solution matrix 
void printSolution(int n, float **dist); 
  
// Solves the all-pairs shortest path problem using Floyd Warshall algorithm 
void floydWarshall_vec(int n, float **graph, float **dist) 
{ 
    // Fill here! 
} 
  

// Solves the all-pairs shortest path problem using Floyd Warshall algorithm 
void floydWarshall(int n, float **graph, float **dist) 
{ 
    /* dist[][] will be the output matrix that will finally have the shortest  
      distances between every pair of vertices */
    int i, j, k; 
  
    /* Initialize the solution matrix same as input graph matrix. Or  
       we can say the initial values of shortest distances are based 
       on shortest paths considering no intermediate vertex. */
    for (i = 0; i < n; i++) 
        for (j = 0; j < n; j++) 
            dist[i][j] = graph[i][j]; 
  
    /* Add all vertices one by one to the set of intermediate vertices. 
      ---> Before start of an iteration, we have shortest distances between all 
      pairs of vertices such that the shortest distances consider only the 
      vertices in set {0, 1, 2, .. k-1} as intermediate vertices. 
      ----> After the end of an iteration, vertex no. k is added to the set of 
      intermediate vertices and the set becomes {0, 1, 2, .. k} */
    for (k = 0; k < n; k++) 
    { 
        // Pick all vertices as source one by one 
        for (i = 0; i < n; i++) 
        { 
            // Pick all vertices as destination for the 
            // above picked source 
            for (j = 0; j < n; j++) 
            { 
                // If vertex k is on the shortest path from 
                // i to j, then update the value of dist[i][j] 
                if (dist[i][k] + dist[k][j] < dist[i][j]) 
                    dist[i][j] = dist[i][k] + dist[k][j]; 
            } 
        } 
    } 
  
    // Print the shortest distance matrix 
    //printSolution(n, dist); 
} 



/* A utility function to print solution */
void printSolution(int n, float **dist) 
{ 
    printf ("The following matrix shows the shortest distances"
            " between every pair of vertices \n"); 
    for (int i = 0; i < n; i++) 
    { 
        for (int j = 0; j < n; j++) 
        { 
            if (dist[i][j] == INF) 
                printf("%7s", "INF"); 
            else
                printf ("%7f", dist[i][j]); 
        } 
        printf("\n"); 
    } 
} 
 
void fill_graph(int n, float **graph)
{   
    for (int i=0; i<n; i++){
        for (int j=0; j<n; j++){
            graph[i][j] = rand() % n;
        }
    }
}
 
int check_dist(int n, float **a, float **b){
    for (int i=0; i<n; i++){
        for (int j=0; j<n; j++){
            if (a[i][j] != b[i][j]) return 0;
        }
    }
    return 1;
}

// allocate a square matrix of floats. Rows are memaligned.
float ** sqm_alloc(int n){
    float ** res = (float **) malloc(sizeof(float *) * n);

    for (int i=0; i<n; i++){
        res[i] = memalign(16, n*sizeof(float));
    }
    return res;
}

// free a square matrix of floats
void sqm_free(int n, float ** sqm){

    for (int i=0; i<n; i++){
        free(sqm[i]);
    }

    free(sqm);
}


// driver program to test above function 
int main(int argc, char * argv[]) 
{ 
    myInt64 t_naive, t_vec;

    if (argc!=2){
        printf("Usage: %s <n>\n", argv[0]);
        exit(1);
    }

    int n = atoi(argv[1]);

    if (n % FLOATS_PER_INSTR != 0){
        printf("Error: n must be a multiple of %i\n", FLOATS_PER_INSTR);
        exit(1);
    }
    
    init_tsc();
    srand(time(NULL));

    float **graph = sqm_alloc(n);
    float **dist = sqm_alloc(n);
    float **dist_vec = sqm_alloc(n);

    fill_graph(n, graph);

    t_naive = start_tsc();
    floydWarshall(n, graph, dist);
    t_naive = stop_tsc(t_naive);
    
    t_vec = start_tsc();
    floydWarshall_vec(n, graph, dist_vec);
    t_vec = stop_tsc(t_vec);
    
    if (check_dist(n, dist, dist_vec)){
        printf("PASSED! Time naive: %llu; Time vec: %llu\n", t_naive, t_vec);
    }else{
        printf("FAILED!\n");
    }

    sqm_free(n, graph);
    sqm_free(n, dist); 
    sqm_free(n, dist_vec);

    return 0; 
} 
